/*
 * Array.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using System.Text.RegularExpressions;
using Eagle._Attributes;
using Eagle._Components.Private;
using Eagle._Components.Public;
using Eagle._Constants;
using Eagle._Containers.Public;
using Eagle._Interfaces.Public;

namespace Eagle._Commands
{
    [ObjectId("fe8bac16-d7a1-4b29-b4a1-7948ee4d9611")]
    [CommandFlags(CommandFlags.Safe | CommandFlags.Standard)]
    [ObjectGroup("variable")]
    internal sealed class Array : Core
    {
        public Array(
            ICommandData commandData
            )
            : base(commandData)
        {
            // do nothing.
        }

        #region IEnsemble Members
        private static readonly EnsembleDictionary subCommands = new EnsembleDictionary(new string[] {
            "anymore", "donesearch", "exists", "get", "names",
            "nextelement", "set", "size", "startsearch", "unset",
            "values"
        });

        public override EnsembleDictionary SubCommands
        {
            get { return subCommands; }
        }
        #endregion

        #region IExecute Members
        public override ReturnCode Execute(
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            ReturnCode code = ReturnCode.Ok;

            if (interpreter != null)
            {
                if (arguments != null)
                {
                    if (arguments.Count >= 2)
                    {
                        string subCommand = arguments[1];
                        bool tried = false;

                        code = RuntimeOps.TryExecuteSubCommandFromEnsemble(
                            this, interpreter, clientData, arguments, true,
                            false, ref subCommand, ref tried, ref result);

                        if ((code == ReturnCode.Ok) && !tried)
                        {
                            switch (subCommand)
                            {
                                case "anymore":
                                    {
                                        if (arguments.Count == 4)
                                        {
                                            lock (interpreter.SyncRoot) /* TRANSACTIONAL */
                                            {
                                                VariableFlags flags = VariableFlags.NoElement |
                                                    VariableFlags.NoLinkIndex | VariableFlags.Defined |
                                                    VariableFlags.NonVirtual;
                                                IVariable variable = null;
                                                Result localError = null;

                                                if (interpreter.GetVariableViaResolversWithSplit(
                                                        arguments[2], ref flags, ref variable,
                                                        ref localError) == ReturnCode.Ok)
                                                {
                                                    if (EntityOps.IsLink(variable))
                                                        variable = EntityOps.FollowLinks(variable);

                                                    if (!EntityOps.IsUndefined(variable) && EntityOps.IsArray(variable))
                                                    {
                                                        if (interpreter.ArraySearches != null)
                                                        {
                                                            ArraySearch arraySearch;

                                                            if (interpreter.ArraySearches.TryGetValue(arguments[3],
                                                                    out arraySearch))
                                                            {
                                                                if (System.Object.ReferenceEquals(arraySearch.Variable, variable))
                                                                {
                                                                    result = arraySearch.AnyMore;
                                                                }
                                                                else
                                                                {
                                                                    result = String.Format(
                                                                        "search identifier \"{0}\" isn't for variable \"{1}\"",
                                                                        arguments[3], arguments[2]);

                                                                    code = ReturnCode.Error;
                                                                }
                                                            }
                                                            else
                                                            {
                                                                result = String.Format(
                                                                    "couldn't find search \"{0}\"",
                                                                    arguments[3]);

                                                                code = ReturnCode.Error;
                                                            }
                                                        }
                                                        else
                                                        {
                                                            result = "array searches not available";
                                                            code = ReturnCode.Error;
                                                        }
                                                    }
                                                    else
                                                    {
                                                        result = String.Format(
                                                            "\"{0}\" isn't an array",
                                                            arguments[2]);

                                                        code = ReturnCode.Error;
                                                    }
                                                }
                                                else
                                                {
                                                    if (FlagOps.HasFlags(flags, VariableFlags.ArrayErrorMask, false))
                                                    {
                                                        result = String.Format(
                                                            "\"{0}\" isn't an array",
                                                            arguments[2]);

                                                        code = ReturnCode.Error;
                                                    }
                                                    else
                                                    {
                                                        result = localError;
                                                        code = ReturnCode.Error;
                                                    }
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"array anymore arrayName searchId\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "donesearch":
                                    {
                                        if (arguments.Count == 4)
                                        {
                                            lock (interpreter.SyncRoot) /* TRANSACTIONAL */
                                            {
                                                VariableFlags flags = VariableFlags.NoElement |
                                                    VariableFlags.NoLinkIndex | VariableFlags.Defined |
                                                    VariableFlags.NonVirtual;
                                                IVariable variable = null;
                                                Result localError = null;

                                                if (interpreter.GetVariableViaResolversWithSplit(
                                                        arguments[2], ref flags, ref variable,
                                                        ref localError) == ReturnCode.Ok)
                                                {
                                                    if (EntityOps.IsLink(variable))
                                                        variable = EntityOps.FollowLinks(variable);

                                                    if (!EntityOps.IsUndefined(variable) && EntityOps.IsArray(variable))
                                                    {
                                                        if (interpreter.ArraySearches != null)
                                                        {
                                                            ArraySearch arraySearch;

                                                            if (interpreter.ArraySearches.TryGetValue(arguments[3],
                                                                    out arraySearch))
                                                            {
                                                                if (System.Object.ReferenceEquals(arraySearch.Variable, variable))
                                                                {
                                                                    interpreter.ArraySearches.Remove(arguments[3]);

                                                                    result = String.Empty;
                                                                }
                                                                else
                                                                {
                                                                    result = String.Format(
                                                                        "search identifier \"{0}\" isn't for variable \"{1}\"",
                                                                        arguments[3], arguments[2]);

                                                                    code = ReturnCode.Error;
                                                                }
                                                            }
                                                            else
                                                            {
                                                                result = String.Format(
                                                                    "couldn't find search \"{0}\"",
                                                                    arguments[3]);

                                                                code = ReturnCode.Error;
                                                            }
                                                        }
                                                        else
                                                        {
                                                            result = "array searches not available";
                                                            code = ReturnCode.Error;
                                                        }
                                                    }
                                                    else
                                                    {
                                                        result = String.Format(
                                                            "\"{0}\" isn't an array",
                                                            arguments[2]);

                                                        code = ReturnCode.Error;
                                                    }
                                                }
                                                else
                                                {
                                                    if (FlagOps.HasFlags(flags, VariableFlags.ArrayErrorMask, false))
                                                    {
                                                        result = String.Format(
                                                            "\"{0}\" isn't an array",
                                                            arguments[2]);

                                                        code = ReturnCode.Error;
                                                    }
                                                    else
                                                    {
                                                        result = localError;
                                                        code = ReturnCode.Error;
                                                    }
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"array donesearch arrayName searchId\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "exists":
                                    {
                                        if (arguments.Count == 3)
                                        {
                                            lock (interpreter.SyncRoot) /* TRANSACTIONAL */
                                            {
                                                VariableFlags flags = VariableFlags.NoElement |
                                                    VariableFlags.NoLinkIndex | VariableFlags.Defined |
                                                    VariableFlags.NonVirtual;
                                                IVariable variable = null;
                                                Result localError = null;

                                                if (interpreter.GetVariableViaResolversWithSplit(
                                                        arguments[2], ref flags, ref variable,
                                                        ref localError) == ReturnCode.Ok)
                                                {
                                                    if (EntityOps.IsLink(variable))
                                                        variable = EntityOps.FollowLinks(variable);

                                                    result = !EntityOps.IsUndefined(variable) &&
                                                        EntityOps.IsArray(variable);
                                                }
                                                else
                                                {
                                                    if (FlagOps.HasFlags(flags, VariableFlags.ArrayErrorMask, false))
                                                    {
                                                        result = false; // variable does not exist.
                                                    }
                                                    else
                                                    {
                                                        result = localError;
                                                        code = ReturnCode.Error;
                                                    }
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"array exists arrayName\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "get":
                                    {
                                        if ((arguments.Count == 3) || (arguments.Count == 4))
                                        {
                                            lock (interpreter.SyncRoot) /* TRANSACTIONAL */
                                            {
                                                VariableFlags flags = VariableFlags.NoElement |
                                                    VariableFlags.NoLinkIndex | VariableFlags.Defined |
                                                    VariableFlags.NonVirtual;
                                                IVariable variable = null;
                                                Result localError = null;

                                                if (interpreter.GetVariableViaResolversWithSplit(
                                                        arguments[2], ref flags, ref variable,
                                                        ref localError) == ReturnCode.Ok)
                                                {
                                                    if (EntityOps.IsLink(variable))
                                                        variable = EntityOps.FollowLinks(variable);

                                                    if (!EntityOps.IsUndefined(variable) && EntityOps.IsArray(variable))
                                                    {
                                                        string pattern = null;

                                                        if (arguments.Count >= 4)
                                                            pattern = arguments[3];

                                                        //
                                                        // HACK: Handle the global "env" array specially.  We must do this because
                                                        //       our global "env" array has no backing storage (unlike Tcl's) and
                                                        //       we do not have a trace operation for "get names" or "get names
                                                        //       and values".
                                                        //
                                                        if (interpreter.IsEnvironmentVariable(variable))
                                                        {
                                                            StringDictionary environment = new StringDictionary(
                                                                Environment.GetEnvironmentVariables());

                                                            result = environment.KeysAndValuesToString(pattern, false);
                                                        }
                                                        else if (interpreter.IsTestsVariable(variable))
                                                        {
                                                            StringDictionary tests = interpreter.GetAllTestInformation(
                                                                ref result);

                                                            if (tests != null)
                                                                result = tests.KeysAndValuesToString(pattern, false);
                                                            else
                                                                code = ReturnCode.Error;
                                                        }
                                                        else
                                                        {
                                                            //
                                                            // FIXME: PRI 4: Variable traces will not be fired here because we are
                                                            //        accessing the array elements via the ArrayValues property and
                                                            //        not through the GetVariableValue method.
                                                            //
                                                            result = variable.ArrayValue.KeysAndValuesToString(pattern, false);
                                                        }
                                                    }
                                                    else
                                                    {
                                                        result = String.Empty;
                                                    }
                                                }
                                                else
                                                {
                                                    if (FlagOps.HasFlags(flags, VariableFlags.ArrayErrorMask, false))
                                                    {
                                                        result = String.Empty;
                                                    }
                                                    else
                                                    {
                                                        result = localError;
                                                        code = ReturnCode.Error;
                                                    }
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"array get arrayName ?pattern?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "names":
                                    {
                                        if ((arguments.Count >= 3) && (arguments.Count <= 5))
                                        {
                                            lock (interpreter.SyncRoot) /* TRANSACTIONAL */
                                            {
                                                VariableFlags flags = VariableFlags.NoElement |
                                                    VariableFlags.NoLinkIndex | VariableFlags.Defined |
                                                    VariableFlags.NonVirtual;
                                                IVariable variable = null;
                                                Result localError = null;

                                                if (interpreter.GetVariableViaResolversWithSplit(
                                                        arguments[2], ref flags, ref variable,
                                                        ref localError) == ReturnCode.Ok)
                                                {
                                                    if (EntityOps.IsLink(variable))
                                                        variable = EntityOps.FollowLinks(variable);

                                                    if (!EntityOps.IsUndefined(variable) && EntityOps.IsArray(variable))
                                                    {
                                                        MatchMode mode = StringOps.DefaultMatchMode;
                                                        string pattern = null;

                                                        switch (arguments.Count)
                                                        {
                                                            case 4:
                                                                pattern = arguments[3];
                                                                break;
                                                            case 5:
                                                                object enumValue = EnumOps.TryParseEnum(
                                                                    typeof(MatchMode), arguments[3], true);

                                                                if (enumValue is MatchMode)
                                                                    mode = (MatchMode)enumValue;
                                                                else
                                                                    mode = MatchMode.None;

                                                                pattern = arguments[4];
                                                                break;
                                                        }

                                                        if (mode != MatchMode.None)
                                                        {
                                                            //
                                                            // HACK: Handle the global "env" array specially.  We must do this because
                                                            //       our global "env" array has no backing storage (unlike Tcl's) and
                                                            //       we do not have a trace operation for "get names" or "get names
                                                            //       and values".
                                                            //
                                                            if (interpreter.IsEnvironmentVariable(variable))
                                                            {
                                                                StringDictionary environment = new StringDictionary(
                                                                    Environment.GetEnvironmentVariables());

                                                                result = environment.KeysToString(
                                                                    mode, pattern, false, RegexOptions.None);
                                                            }
                                                            else if (interpreter.IsTestsVariable(variable))
                                                            {
                                                                StringDictionary tests = interpreter.GetAllTestInformation(
                                                                    ref result);

                                                                if (tests != null)
                                                                    result = tests.KeysToString(
                                                                        mode, pattern, false, RegexOptions.None);
                                                                else
                                                                    code = ReturnCode.Error;
                                                            }
                                                            else
                                                            {
                                                                //
                                                                // FIXME: PRI 4: Variable traces will not be fired here because we are
                                                                //        accessing the array elements via the ArrayValues property and
                                                                //        not through the GetVariableValue method.
                                                                //
                                                                result = variable.ArrayValue.KeysToString(
                                                                    mode, pattern, false, RegexOptions.None);
                                                            }
                                                        }
                                                        else
                                                        {
                                                            result = String.Format(
                                                                "bad option \"{0}\": must be -exact, -substring, -glob, or -regexp",
                                                                arguments[3]);

                                                            code = ReturnCode.Error;
                                                        }
                                                    }
                                                    else
                                                    {
                                                        result = String.Empty;
                                                    }
                                                }
                                                else
                                                {
                                                    if (FlagOps.HasFlags(flags, VariableFlags.ArrayErrorMask, false))
                                                    {
                                                        result = String.Empty;
                                                    }
                                                    else
                                                    {
                                                        result = localError;
                                                        code = ReturnCode.Error;
                                                    }
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"array names arrayName ?mode? ?pattern?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "nextelement":
                                    {
                                        if (arguments.Count == 4)
                                        {
                                            lock (interpreter.SyncRoot) /* TRANSACTIONAL */
                                            {
                                                VariableFlags flags = VariableFlags.NoElement |
                                                    VariableFlags.NoLinkIndex | VariableFlags.Defined |
                                                    VariableFlags.NonVirtual;
                                                IVariable variable = null;
                                                Result localError = null;

                                                if (interpreter.GetVariableViaResolversWithSplit(
                                                        arguments[2], ref flags, ref variable,
                                                        ref localError) == ReturnCode.Ok)
                                                {
                                                    if (EntityOps.IsLink(variable))
                                                        variable = EntityOps.FollowLinks(variable);

                                                    if (!EntityOps.IsUndefined(variable) && EntityOps.IsArray(variable))
                                                    {
                                                        if (interpreter.ArraySearches != null)
                                                        {
                                                            ArraySearch arraySearch;

                                                            if (interpreter.ArraySearches.TryGetValue(arguments[3],
                                                                    out arraySearch))
                                                            {
                                                                if (System.Object.ReferenceEquals(arraySearch.Variable, variable))
                                                                {
                                                                    result = arraySearch.GetNextElement();
                                                                }
                                                                else
                                                                {
                                                                    result = String.Format(
                                                                        "search identifier \"{0}\" isn't for variable \"{1}\"",
                                                                        arguments[3], arguments[2]);

                                                                    code = ReturnCode.Error;
                                                                }
                                                            }
                                                            else
                                                            {
                                                                result = String.Format(
                                                                    "couldn't find search \"{0}\"",
                                                                    arguments[3]);

                                                                code = ReturnCode.Error;
                                                            }
                                                        }
                                                        else
                                                        {
                                                            result = "array searches not available";
                                                            code = ReturnCode.Error;
                                                        }
                                                    }
                                                    else
                                                    {
                                                        result = String.Format(
                                                            "\"{0}\" isn't an array",
                                                            arguments[2]);

                                                        code = ReturnCode.Error;
                                                    }
                                                }
                                                else
                                                {
                                                    if (FlagOps.HasFlags(flags, VariableFlags.ArrayErrorMask, false))
                                                    {
                                                        result = String.Format(
                                                            "\"{0}\" isn't an array",
                                                            arguments[2]);

                                                        code = ReturnCode.Error;
                                                    }
                                                    else
                                                    {
                                                        result = localError;
                                                        code = ReturnCode.Error;
                                                    }
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"array nextelement arrayName searchId\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "set":
                                    {
                                        if (arguments.Count == 4)
                                        {
                                            lock (interpreter.SyncRoot) /* TRANSACTIONAL */
                                            {
                                                //
                                                // BUGFIX: If we are not going to add a new array variable, the least
                                                //         we need to do is validate that the argument refers to an
                                                //         existing array and that it does not refer to an element of
                                                //         that array.
                                                //
                                                VariableFlags flags = VariableFlags.Array | VariableFlags.NoElement |
                                                    VariableFlags.NoGetArray | VariableFlags.NoLinkIndex |
                                                    VariableFlags.Defined | VariableFlags.NonVirtual;
                                                IVariable variable = null;
                                                Result localError = null;

                                                if ((interpreter.GetVariableViaResolversWithSplit(
                                                        arguments[2], ref flags, ref variable,
                                                        ref localError) == ReturnCode.Ok) ||
                                                    (interpreter.AddVariable2(
                                                        flags & ~VariableFlags.Defined, arguments[2],
                                                        null, true, ref variable,
                                                        ref localError) == ReturnCode.Ok))
                                                {
                                                    if (EntityOps.IsLink(variable))
                                                        variable = EntityOps.FollowLinks(variable);

                                                    if (!interpreter.IsEnvironmentVariable(variable) &&
                                                        !interpreter.IsTestsVariable(variable))
                                                    {
                                                        if (EntityOps.IsUndefined(variable) ||
                                                            EntityOps.IsArray2(variable))
                                                        {
                                                            StringList list = null;

                                                            if (Parser.SplitList(
                                                                    interpreter, arguments[3], 0,
                                                                    Length.Invalid, false, true,
                                                                    ref list, ref result) == ReturnCode.Ok)
                                                            {
                                                                if ((list.Count & 1) == 0)
                                                                {
                                                                    ObjectDictionary arrayValue =
                                                                        new ObjectDictionary(list.Count / 2);

                                                                    for (int index = 0; index < list.Count; index += 2)
                                                                        arrayValue[list[index]] = list[index + 1];

                                                                    //
                                                                    // BUGFIX: We *MUST* fire traces here; otherwise, both
                                                                    //         old and new opaque object handles could have
                                                                    //         incorrect reference counts.
                                                                    //
                                                                    code = interpreter.FireTraces(
                                                                        BreakpointType.BeforeVariableSet, flags,
                                                                        variable.Frame, arguments[2], null, null,
                                                                        variable.ArrayValue, arrayValue, variable,
                                                                        ref result);

                                                                    if (code == ReturnCode.Ok)
                                                                    {
                                                                        //
                                                                        // HACK: Reset the array value for this variable to
                                                                        //       the one we just created, bypassing the
                                                                        //       entire SetVariableValue* subsystem (for
                                                                        //       speed).
                                                                        //
                                                                        variable.ArrayValue = arrayValue;

                                                                        //
                                                                        // BUGFIX: The variable is now defined.  Mark the
                                                                        //         variable as "dirty" AFTER all the actual
                                                                        //         modifications have been completed.
                                                                        //
                                                                        EntityOps.SetArray(variable, true);
                                                                        EntityOps.SetUndefined(variable, false);
                                                                        EntityOps.SetDirty(variable, true);

                                                                        //
                                                                        // NOTE: This command returns an empty result on
                                                                        //       success.
                                                                        //
                                                                        result = String.Empty;
                                                                    }
                                                                }
                                                                else
                                                                {
                                                                    result = "list must have an even number of elements";
                                                                    code = ReturnCode.Error;
                                                                }
                                                            }
                                                            else
                                                            {
                                                                code = ReturnCode.Error;
                                                            }
                                                        }
                                                        else
                                                        {
                                                            result = String.Format(
                                                                "\"{0}\" isn't an array",
                                                                arguments[2]);

                                                            code = ReturnCode.Error;
                                                        }
                                                    }
                                                    else
                                                    {
                                                        result = String.Format(
                                                            "\"{0}\" is system array",
                                                            arguments[2]);

                                                        code = ReturnCode.Error;
                                                    }
                                                }
                                                else
                                                {
                                                    //
                                                    // TODO: Is this check right?  If so, needs test cases.
                                                    //
                                                    if (!FlagOps.HasFlags(flags, VariableFlags.NotFound, true) ||
                                                        FlagOps.HasFlags(flags, VariableFlags.HasLinkIndex, true))
                                                    {
                                                        result = String.Format(
                                                            "\"{0}\" isn't an array",
                                                            arguments[2]);

                                                        code = ReturnCode.Error;
                                                    }
                                                    else
                                                    {
                                                        result = localError;
                                                        code = ReturnCode.Error;
                                                    }
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"array set arrayName list\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "size":
                                    {
                                        if (arguments.Count == 3)
                                        {
                                            lock (interpreter.SyncRoot) /* TRANSACTIONAL */
                                            {
                                                VariableFlags flags = VariableFlags.NoElement |
                                                    VariableFlags.NoLinkIndex | VariableFlags.Defined |
                                                    VariableFlags.NonVirtual;
                                                IVariable variable = null;
                                                Result localError = null;

                                                if (interpreter.GetVariableViaResolversWithSplit(
                                                        arguments[2], ref flags, ref variable,
                                                        ref localError) == ReturnCode.Ok)
                                                {
                                                    if (EntityOps.IsLink(variable))
                                                        variable = EntityOps.FollowLinks(variable);

                                                    if (!EntityOps.IsUndefined(variable) && EntityOps.IsArray(variable))
                                                    {
                                                        //
                                                        // HACK: Handle the global "env" array specially.  We must do this because
                                                        //       our global "env" array has no backing storage (unlike Tcl's) and
                                                        //       we do not have a trace operation for "get names" or "get names
                                                        //       and values".
                                                        //
                                                        if (interpreter.IsEnvironmentVariable(variable))
                                                        {
                                                            StringDictionary environment = new StringDictionary(
                                                                Environment.GetEnvironmentVariables());

                                                            result = environment.Count;
                                                        }
                                                        else if (interpreter.IsTestsVariable(variable))
                                                        {
                                                            StringDictionary tests = interpreter.GetAllTestInformation(
                                                                ref result);

                                                            if (tests != null)
                                                                result = tests.Count;
                                                            else
                                                                code = ReturnCode.Error;
                                                        }
                                                        else
                                                        {
                                                            //
                                                            // FIXME: PRI 4: Variable traces will not be fired here because we are
                                                            //        accessing the array elements via the ArrayValues property and
                                                            //        not through the GetVariableValue method.
                                                            //
                                                            result = variable.ArrayValue.Count;
                                                        }
                                                    }
                                                    else
                                                    {
                                                        result = 0; // variable is not array.
                                                    }
                                                }
                                                else
                                                {
                                                    if (FlagOps.HasFlags(flags, VariableFlags.ArrayErrorMask, false))
                                                    {
                                                        result = 0; // variable does not exist.
                                                    }
                                                    else
                                                    {
                                                        result = localError;
                                                        code = ReturnCode.Error;
                                                    }
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"array size arrayName\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "startsearch":
                                    {
                                        if (arguments.Count == 3)
                                        {
                                            lock (interpreter.SyncRoot) /* TRANSACTIONAL */
                                            {
                                                VariableFlags flags = VariableFlags.NoElement |
                                                    VariableFlags.NoLinkIndex | VariableFlags.Defined |
                                                    VariableFlags.NonVirtual;
                                                IVariable variable = null;
                                                Result localError = null;

                                                if (interpreter.GetVariableViaResolversWithSplit(
                                                        arguments[2], ref flags, ref variable,
                                                        ref localError) == ReturnCode.Ok)
                                                {
                                                    if (EntityOps.IsLink(variable))
                                                        variable = EntityOps.FollowLinks(variable);

                                                    if (!EntityOps.IsUndefined(variable) && EntityOps.IsArray(variable))
                                                    {
                                                        if (interpreter.ArraySearches != null)
                                                        {
                                                            ArraySearch arraySearch = new ArraySearch(interpreter, variable);
                                                            string name = FormatOps.Id("arraySearch", null, interpreter.NextId());

                                                            //
                                                            // NOTE: Add the new array search to the interpreter.
                                                            //
                                                            interpreter.ArraySearches.Add(name, arraySearch);

                                                            //
                                                            // NOTE: Return the Id of the newly created array search
                                                            //       to the caller.
                                                            //
                                                            result = name;
                                                        }
                                                        else
                                                        {
                                                            result = "array searches not available";
                                                            code = ReturnCode.Error;
                                                        }
                                                    }
                                                    else
                                                    {
                                                        result = String.Format(
                                                            "\"{0}\" isn't an array",
                                                            arguments[2]);

                                                        code = ReturnCode.Error;
                                                    }
                                                }
                                                else
                                                {
                                                    if (FlagOps.HasFlags(flags, VariableFlags.ArrayErrorMask, false))
                                                    {
                                                        result = String.Format(
                                                            "\"{0}\" isn't an array",
                                                            arguments[2]);

                                                        code = ReturnCode.Error;
                                                    }
                                                    else
                                                    {
                                                        result = localError;
                                                        code = ReturnCode.Error;
                                                    }
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"array startsearch arrayName\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "unset":
                                    {
                                        if ((arguments.Count == 3) || (arguments.Count == 4))
                                        {
                                            lock (interpreter.SyncRoot) /* TRANSACTIONAL */
                                            {
                                                VariableFlags flags = VariableFlags.NoElement |
                                                    VariableFlags.NoLinkIndex | VariableFlags.Defined |
                                                    VariableFlags.NonVirtual;
                                                IVariable variable = null;
                                                Result localError = null;

                                                if (interpreter.GetVariableViaResolversWithSplit(
                                                        arguments[2], ref flags, ref variable,
                                                        ref localError) == ReturnCode.Ok)
                                                {
                                                    if (EntityOps.IsLink(variable))
                                                        variable = EntityOps.FollowLinks(variable);

                                                    if (!EntityOps.IsUndefined(variable) && EntityOps.IsArray(variable))
                                                    {
                                                        if (arguments.Count == 4)
                                                        {
                                                            StringList list = null;

                                                            //
                                                            // HACK: Handle the global "env" array specially.  We must do this because
                                                            //       our global "env" array has no backing storage (unlike Tcl's) and
                                                            //       we do not have a trace operation for "get names" or "get names
                                                            //       and values".
                                                            //
                                                            if (interpreter.IsEnvironmentVariable(variable))
                                                            {
                                                                StringDictionary environment = new StringDictionary(
                                                                    Environment.GetEnvironmentVariables());

                                                                list = new StringList(environment.Keys);
                                                            }
                                                            else if (interpreter.IsTestsVariable(variable))
                                                            {
                                                                StringDictionary tests = interpreter.GetAllTestInformation(
                                                                    ref result);

                                                                if (tests != null)
                                                                    list = new StringList(tests.Keys);
                                                                else
                                                                    code = ReturnCode.Error;
                                                            }
                                                            else
                                                            {
                                                                //
                                                                // FIXME: PRI 4: Variable traces will not be fired here because we are
                                                                //        accessing the array elements via the ArrayValues property and
                                                                //        not through the GetVariableValue method.
                                                                //
                                                                list = new StringList(variable.ArrayValue.Keys);
                                                            }

                                                            //
                                                            // NOTE: Is there anything to do (i.e. elements to unset)?
                                                            //
                                                            if (list != null)
                                                            {
                                                                foreach (string element in list)
                                                                {
                                                                    if (Parser.StringMatch(interpreter, element, 0,
                                                                            arguments[3], 0, false))
                                                                    {
                                                                        code = interpreter.UnsetVariable2(
                                                                            VariableFlags.None, arguments[2],
                                                                            element, variable, ref result);

                                                                        if (code != ReturnCode.Ok)
                                                                            break;
                                                                    }
                                                                }

                                                                if (code == ReturnCode.Ok)
                                                                    result = String.Empty;
                                                            }
                                                        }
                                                        else
                                                        {
                                                            //
                                                            // NOTE: Remove the entire array.
                                                            //
                                                            code = interpreter.UnsetVariable2(VariableFlags.None, arguments[2],
                                                                null, variable, ref result);
                                                        }
                                                    }
                                                }
                                                else
                                                {
                                                    if (FlagOps.HasFlags(flags, VariableFlags.ArrayErrorMask, false))
                                                    {
                                                        result = String.Empty;
                                                    }
                                                    else
                                                    {
                                                        result = localError;
                                                        code = ReturnCode.Error;
                                                    }
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"array unset arrayName ?pattern?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                case "values":
                                    {
                                        if ((arguments.Count >= 3) && (arguments.Count <= 5))
                                        {
                                            lock (interpreter.SyncRoot) /* TRANSACTIONAL */
                                            {
                                                VariableFlags flags = VariableFlags.NoElement |
                                                    VariableFlags.NoLinkIndex | VariableFlags.Defined |
                                                    VariableFlags.NonVirtual;
                                                IVariable variable = null;
                                                Result localError = null;

                                                if (interpreter.GetVariableViaResolversWithSplit(
                                                        arguments[2], ref flags, ref variable,
                                                        ref localError) == ReturnCode.Ok)
                                                {
                                                    if (EntityOps.IsLink(variable))
                                                        variable = EntityOps.FollowLinks(variable);

                                                    if (!EntityOps.IsUndefined(variable) && EntityOps.IsArray(variable))
                                                    {
                                                        MatchMode mode = StringOps.DefaultMatchMode;
                                                        string pattern = null;

                                                        switch (arguments.Count)
                                                        {
                                                            case 4:
                                                                pattern = arguments[3];
                                                                break;
                                                            case 5:
                                                                object enumValue = EnumOps.TryParseEnum(
                                                                    typeof(MatchMode), arguments[3], true);

                                                                if (enumValue is MatchMode)
                                                                    mode = (MatchMode)enumValue;
                                                                else
                                                                    mode = MatchMode.None;

                                                                pattern = arguments[4];
                                                                break;
                                                        }

                                                        if (mode != MatchMode.None)
                                                        {
                                                            //
                                                            // HACK: Handle the global "env" array specially.  We must do this because
                                                            //       our global "env" array has no backing storage (unlike Tcl's) and
                                                            //       we do not have a trace operation for "get names" or "get names
                                                            //       and values".
                                                            //
                                                            if (interpreter.IsEnvironmentVariable(variable))
                                                            {
                                                                StringDictionary environment = new StringDictionary(
                                                                    Environment.GetEnvironmentVariables());

                                                                result = environment.ValuesToString(
                                                                    mode, pattern, false, RegexOptions.None);
                                                            }
                                                            else if (interpreter.IsTestsVariable(variable))
                                                            {
                                                                StringDictionary tests = interpreter.GetAllTestInformation(
                                                                    ref result);

                                                                if (tests != null)
                                                                    result = tests.ValuesToString(
                                                                        mode, pattern, false, RegexOptions.None);
                                                                else
                                                                    code = ReturnCode.Error;
                                                            }
                                                            else
                                                            {
                                                                //
                                                                // FIXME: PRI 4: Variable traces will not be fired here because we are
                                                                //        accessing the array elements via the ArrayValues property and
                                                                //        not through the GetVariableValue method.
                                                                //
                                                                result = variable.ArrayValue.ValuesToString(
                                                                    mode, pattern, false, RegexOptions.None);
                                                            }
                                                        }
                                                        else
                                                        {
                                                            result = String.Format(
                                                                "bad option \"{0}\": must be -exact, -substring, -glob, or -regexp",
                                                                arguments[3]);

                                                            code = ReturnCode.Error;
                                                        }
                                                    }
                                                    else
                                                    {
                                                        result = String.Empty;
                                                    }
                                                }
                                                else
                                                {
                                                    if (FlagOps.HasFlags(flags, VariableFlags.ArrayErrorMask, false))
                                                    {
                                                        result = String.Empty;
                                                    }
                                                    else
                                                    {
                                                        result = localError;
                                                        code = ReturnCode.Error;
                                                    }
                                                }
                                            }
                                        }
                                        else
                                        {
                                            result = "wrong # args: should be \"array values arrayName ?mode? ?pattern?\"";
                                            code = ReturnCode.Error;
                                        }
                                        break;
                                    }
                                default:
                                    {
                                        result = RuntimeOps.BadSubCommand(null, null, subCommand, this, null, null);
                                        code = ReturnCode.Error;
                                        break;
                                    }
                            }
                        }
                    }
                    else
                    {
                        result = "wrong # args: should be \"array option ?arg ...?\"";
                        code = ReturnCode.Error;
                    }
                }
                else
                {
                    result = "invalid argument list";
                    code = ReturnCode.Error;
                }
            }
            else
            {
                result = "invalid interpreter";
                code = ReturnCode.Error;
            }

            return code;
        }
        #endregion
    }
}
