/*
 * Uplevel.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using Eagle._Attributes;
using Eagle._Components.Private;
using Eagle._Components.Public;
using Eagle._Containers.Public;
using Eagle._Interfaces.Public;

namespace Eagle._Commands
{
    [ObjectId("bafa3e2b-b26c-4552-9365-f9089d757e33")]
    [CommandFlags(CommandFlags.Safe | CommandFlags.Standard)]
    [ObjectGroup("control")]
    internal sealed class Uplevel : Core
    {
        public Uplevel(
            ICommandData commandData
            )
            : base(commandData)
        {
            // do nothing.
        }

        #region IExecute Members
        public override ReturnCode Execute(
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            ReturnCode code;

            if (interpreter != null)
            {
                if (arguments != null)
                {
                    if (arguments.Count >= 2)
                    {
                        bool mark = false;
                        bool absolute = false;
                        int level = 0;
                        ICallFrame otherFrame = null;

                        FrameResult frameResult = Value.GetCallFrame(
                            arguments[1], LevelFlags.All, interpreter.CallStack, interpreter.GlobalFrame,
                            interpreter.CurrentFrame, CallFrameFlags.Variables, CallFrameFlags.Invisible,
                            false, false, interpreter.CultureInfo, false, ref mark, ref absolute, ref level,
                            ref otherFrame, ref result);

                        if (frameResult != FrameResult.Invalid)
                        {
                            int argumentIndex = ((int)frameResult + 1);

                            //
                            // BUGFIX: The argument count needs to be checked again here.
                            //
                            if (argumentIndex < arguments.Count)
                            {
                                if (mark)
                                    code = CallFrameOps.MarkMatching(
                                        interpreter.CallStack, interpreter.CurrentFrame, absolute,
                                        level, CallFrameFlags.Variables,
                                        CallFrameFlags.Invisible | CallFrameFlags.Tracking,
                                        CallFrameFlags.Invisible, false, false, true, ref result);
                                else
                                    code = ReturnCode.Ok;

                                if (code == ReturnCode.Ok)
                                {
                                    string name = StringList.MakeList("uplevel", arguments[1]);

                                    //
                                    // HACK: If we are not marking call frames as "invisible" then 
                                    //       we must be in "view everything" mode; therefore, set
                                    //       the NoInvisible flag on this call frame so that it 
                                    //       will be found by [info level], etc.
                                    //
                                    ICallFrame frame = interpreter.NewLinkedCallFrame(
                                        name, null, CallFrameFlags.Uplevel |
                                            (mark ? CallFrameFlags.None : CallFrameFlags.NoInvisible),
                                        otherFrame);

                                    interpreter.PushCallFrame(frame);

                                    if ((argumentIndex + 1) >= arguments.Count)
                                        code = interpreter.EvaluateScript(
                                            arguments[argumentIndex], ref result);
                                    else
                                        code = interpreter.EvaluateScript(
                                            arguments, argumentIndex, ref result);

                                    if (code == ReturnCode.Error)
                                        Engine.AddErrorInformation(interpreter, result,
                                            String.Format("{0}    (\"uplevel\" body line {1})",
                                                Environment.NewLine, interpreter.ErrorLine));

                                    //
                                    // NOTE: Pop the original call frame that we pushed above and 
                                    //       any intervening scope call frames that may be leftover 
                                    //       (i.e. they were not explicitly closed).
                                    //
                                    interpreter.PopCallFrame(CallFrameFlags.Scope, true);

                                    if (mark)
                                    {
                                        //
                                        // NOTE: We should not get an error at this point from unmarking
                                        //       the call frames; however, if we do get one, we need to 
                                        //       complain loudly about it because that means the interpreter 
                                        //       state has probably been corrupted somehow.
                                        //
                                        ReturnCode markCode;
                                        Result markResult = null;

                                        markCode = CallFrameOps.MarkMatching(
                                            interpreter.CallStack, interpreter.CurrentFrame, absolute, level,
                                            CallFrameFlags.Variables, CallFrameFlags.Tracking,
                                            CallFrameFlags.Invisible, false, false, false, ref markResult);

                                        if (markCode != ReturnCode.Ok)
                                            DebugOps.Complain(interpreter, markCode, markResult);
                                    }
                                }
                            }
                            else
                            {
                                result = "wrong # args: should be \"uplevel ?level? arg ?arg ...?\"";
                                code = ReturnCode.Error;
                            }
                        }
                        else
                        {
                            code = ReturnCode.Error;
                        }
                    }
                    else
                    {
                        result = "wrong # args: should be \"uplevel ?level? arg ?arg ...?\"";
                        code = ReturnCode.Error;
                    }
                }
                else
                {
                    result = "invalid argument list";
                    code = ReturnCode.Error;
                }
            }
            else
            {
                result = "invalid interpreter";
                code = ReturnCode.Error;
            }

            return code;
        }
        #endregion        
    }
}
