/*
 * Variable.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using Eagle._Attributes;
using Eagle._Components.Private;
using Eagle._Components.Public;
using Eagle._Containers.Public;
using Eagle._Interfaces.Public;

namespace Eagle._Commands
{
    [ObjectId("8f887079-44e3-405a-a0d3-0b446ce2fa15")]
    [CommandFlags(CommandFlags.Safe | CommandFlags.Standard)]
    [ObjectGroup("variable")]
    internal sealed class _Variable : Core
    {
        public _Variable(
            ICommandData commandData
            )
            : base(commandData)
        {
            // do nothing.
        }

        #region IExecute Members
        public override ReturnCode Execute(
            Interpreter interpreter,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            ReturnCode code = ReturnCode.Ok;

            if (interpreter != null)
            {
                if (arguments != null)
                {
                    if (arguments.Count >= 2)
                    {
                        for (int argumentIndex = 1; argumentIndex < arguments.Count; argumentIndex += 2)
                        {
                            lock (interpreter.SyncRoot) /* TRANSACTIONAL */
                            {
                                VariableFlags flags = VariableFlags.GlobalOnly | VariableFlags.NoElement;
                                IVariable variable = null;
                                Result localError = null;

                                code = interpreter.GetVariableViaResolversWithSplit(
                                    arguments[argumentIndex], ref flags, ref variable,
                                    ref localError);

                                if (code != ReturnCode.Ok)
                                {
                                    if (FlagOps.HasFlags(flags, VariableFlags.NotFound, true))
                                    {
                                        code = interpreter.AddVariable2(
                                            VariableFlags.Undefined | flags,
                                            arguments[argumentIndex], null, true,
                                            ref variable, ref result);
                                    }
                                    else
                                    {
                                        //
                                        // NOTE: We did not search for the variable, let
                                        //       the caller know why.
                                        //
                                        result = localError;
                                    }
                                }

                                //
                                // NOTE: Did we manage to search for (and find) the
                                //       variable or add the new variable?
                                //
                                if (code == ReturnCode.Ok)
                                {
                                    if ((argumentIndex + 1) < arguments.Count)
                                    {
                                        //
                                        // NOTE: They provided a value, set it now.
                                        //
                                        code = interpreter.SetVariableValue(
                                            VariableFlags.None, arguments[argumentIndex],
                                            arguments[argumentIndex + 1], null,
                                            ref variable, ref result);

                                        if (code != ReturnCode.Ok)
                                            break;
                                    }

                                    //
                                    // NOTE: Are we NOT in the global call frame?
                                    //
                                    ICallFrame currentFrame =
                                        CallFrameOps.FollowLinks(interpreter.CurrentFrame);

                                    if ((currentFrame != null) &&
                                        !interpreter.IsGlobalCallFrame(currentFrame))
                                    {
                                        code = RuntimeOps.LinkVariable(
                                            interpreter, currentFrame,
                                            arguments[argumentIndex], interpreter.GlobalFrame,
                                            arguments[argumentIndex], ref result);

                                        if (code != ReturnCode.Ok)
                                            break;
                                    }
                                }
                                else
                                {
                                    break;
                                }
                            }
                        }

                        if (code == ReturnCode.Ok)
                            result = String.Empty;
                    }
                    else
                    {
                        result = "wrong # args: should be \"variable ?name value...? name ?value?\"";
                        code = ReturnCode.Error;
                    }
                }
                else
                {
                    result = "invalid argument list";
                    code = ReturnCode.Error;
                }
            }
            else
            {
                result = "invalid interpreter";
                code = ReturnCode.Error;
            }

            return code;
        }
        #endregion
    }
}
