/*
 * PerformanceOps.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

#if !NATIVE || !WINDOWS
using System;
#endif

using System.Runtime.InteropServices;

#if NATIVE
using System.Security;
#endif

#if !NET_40
using System.Security.Permissions;
#endif

using Eagle._Attributes;
using Eagle._Constants;

namespace Eagle._Components.Private
{
#if NATIVE
#if NET_40
    [SecurityCritical()]
#else
    [SecurityPermission(SecurityAction.LinkDemand, UnmanagedCode = true)]
#endif
#endif
    [ObjectId("4e5f0e33-37e0-4d99-a39f-6d0a87ebc487")]
    internal static class PerformanceOps
    {
#if NATIVE
        ///////////////////////////////////////////////////////////////////////////////////////////////
        // Required Native APIs used via P/Invoke
        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private Unsafe Native Methods Class
        [SuppressUnmanagedCodeSecurity()]
        [ObjectId("53eaca2a-6ad1-4373-b541-8decca686521")]
        private static class UnsafeNativeMethods
        {
#if WINDOWS
            [DllImport(DllName.Kernel32, CallingConvention = CallingConvention.Winapi, SetLastError = true)]
            [return: MarshalAs(UnmanagedType.Bool)]
            internal static extern bool QueryPerformanceCounter(out long count);

            [DllImport(DllName.Kernel32, CallingConvention = CallingConvention.Winapi, SetLastError = true)]
            [return: MarshalAs(UnmanagedType.Bool)]
            internal static extern bool QueryPerformanceFrequency(out long frequency);
#endif
        }
        #endregion
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if !NATIVE || !WINDOWS
        private const long MillisecondsPerSecond = 1000;
#endif

        private const long MicrosecondsPerMillisecond = 1000;
        private const long MicrosecondsPerSecond = 1000000;

        private static readonly long CountsPerSecond = GetCountsPerSecond();

        #region Dead Code
#if DEAD_CODE
#if !NATIVE || !WINDOWS
        private static readonly double CountsPerMillisecond =
            (double)CountsPerSecond / (double)MillisecondsPerSecond;
#endif
#endif
        #endregion

        private static readonly double CountsPerMicrosecond =
            (double)CountsPerSecond / (double)MicrosecondsPerSecond;

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static long GetMicroseconds(
            long milliseconds
            )
        {
            return milliseconds * MicrosecondsPerMillisecond;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static double GetMicroseconds(
            long startCount,
            long stopCount,
            long iterations
            )
        {
            double result;

            if (iterations <= 1)
                result = (stopCount - startCount) / CountsPerMicrosecond;
            else
                result = ((stopCount - startCount) / (double)iterations) / CountsPerMicrosecond;

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static long GetMilliseconds(
            long microseconds
            )
        {
            return microseconds / MicrosecondsPerMillisecond;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Dead Code
#if DEAD_CODE
#if !NATIVE || !WINDOWS
        private static double GetMilliseconds(
            long startCount,
            long stopCount,
            long iterations
            )
        {
            double result;

            if (iterations <= 1)
                result = (stopCount - startCount) / CountsPerMillisecond;
            else
                result = ((stopCount - startCount) / (double)iterations) / CountsPerMillisecond;

            return result;
        }
#endif
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static long GetCountsPerSecond()
        {
            long result;

#if NATIVE && WINDOWS
            if (!UnsafeNativeMethods.QueryPerformanceFrequency(out result))
                result = 0;
#else
            result = MicrosecondsPerSecond;
#endif

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static long GetCount()
        {
            long result;

#if NATIVE && WINDOWS
            if (!UnsafeNativeMethods.QueryPerformanceCounter(out result))
                result = 0;
#else
            //
            // BUGFIX: This result must be in microseconds because the various
            //         callers of this method assume that is the case when they
            //         calculate elapsed time.
            //
            result = Environment.TickCount * MicrosecondsPerMillisecond;
#endif

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static double ElapsedMicroseconds(
            long startCount
            )
        {
            return GetMicroseconds(startCount, GetCount(), 1);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static bool HasElapsed(
            long startCount,
            long waitMicroseconds,
            long slopMicroseconds
            )
        {
            return ElapsedMicroseconds(startCount) +
                (double)slopMicroseconds >= (double)waitMicroseconds;
        }
    }
}
