/*
 * Object.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System.Collections.Generic;
using Eagle._Attributes;
using Eagle._Components.Private;
using Eagle._Components.Public;
using Eagle._Constants;
using Eagle._Containers.Public;
using Eagle._Interfaces.Public;

namespace Eagle._Plugins
{
    [ObjectId("55febaed-b731-4d2c-9176-4144a7550011")]
    [PluginFlags(
        PluginFlags.System | PluginFlags.Notify |
        PluginFlags.Static | PluginFlags.NoCommands |
        PluginFlags.NoFunctions | PluginFlags.NoPolicies |
        PluginFlags.NoTraces)]
    [NotifyTypes(NotifyType.CallFrame)]
    [NotifyFlags(NotifyFlags.Popped | NotifyFlags.Deleted)]
    internal sealed class Object : Notify
    {
        #region Public Constructors
        public Object(
            IPluginData pluginData
            )
            : base(pluginData)
        {
            this.Flags |= AttributeOps.GetPluginFlags(GetType().BaseType) |
                AttributeOps.GetPluginFlags(this);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region IState Members
        public override ReturnCode Terminate(
            Interpreter interpreter,
            IClientData clientData,
            ref Result result
            )
        {
            //
            // HACK: Cleanup the object plugin token in the interpreter
            //       state because this is the only place we can be
            //       100% sure it will get done.
            //
            if (interpreter != null)
                interpreter.objectPluginToken = 0;

            return base.Terminate(interpreter, clientData, ref result);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region INotify Members
        public override ReturnCode Notify(
            Interpreter interpreter,
            IScriptEventArgs eventArgs,
            IClientData clientData,
            ArgumentList arguments,
            ref Result result
            )
        {
            ReturnCode code = ReturnCode.Ok;

            if (eventArgs != null)
            {
                if (FlagOps.HasFlags(eventArgs.Types,
                        NotifyType.CallFrame, false) &&
                    FlagOps.HasFlags(eventArgs.Flags,
                        NotifyFlags.Popped | NotifyFlags.Deleted, false))
                {
                    Interpreter eventInterpreter = eventArgs.Interpreter;

                    if ((eventInterpreter != null) && (eventArgs.ClientData != null))
                    {
                        ICallFrame newFrame = eventArgs.ClientData.Data as ICallFrame;

                        if (newFrame != null)
                        {
                            //
                            // NOTE: Make sure the variables in this frame actually BELONG
                            //       to this frame.  Also, we do not handle the global call
                            //       frame.
                            //
                            if (CallFrameOps.IsVariable(newFrame) &&
                                !CallFrameOps.IsGlobal(newFrame))
                            {
                                //
                                // NOTE: If this is a [scope] created call frame, we do NOT want
                                //       to change any reference counts unless the call frame is
                                //       being deleted, not simply popped.
                                //
                                if (!CallFrameOps.IsScope(newFrame) ||
                                    FlagOps.HasFlags(eventArgs.Flags, NotifyFlags.Deleted, true))
                                {
                                    //
                                    // NOTE: Process each variable in the call frame to properly
                                    //       cleanup all reference counts.
                                    //
                                    foreach (KeyValuePair<string, IVariable> pair in newFrame.Variables)
                                    {
                                        //
                                        // NOTE: Make sure the variable is valid.
                                        //
                                        if (pair.Value != null)
                                        {
                                            //
                                            // NOTE: For unset, ObjectTraceCallback uses only the
                                            //       traceInfo.Variable and traceInfo.oldValue
                                            //       members.
                                            //
                                            ITraceInfo traceInfo = RuntimeOps.NewTraceInfo(interpreter,
                                                null, BreakpointType.BeforeVariableUnset, newFrame,
                                                pair.Value, pair.Key, null, VariableFlags.None,
                                                pair.Value.Value, null, null, null, null, false, true,
                                                ReturnCode.Ok);

                                            //
                                            // HACK: Notify the trace callback that the object
                                            //       reference is going away now.
                                            //
                                            code = Interpreter.ObjectTraceCallback(
                                                traceInfo.BreakpointType, eventInterpreter, traceInfo,
                                                ref result);

                                            if (code != ReturnCode.Ok)
                                                break;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }

            return code;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////

        #region IPlugin Members
        public override ReturnCode About(
            Interpreter interpreter,
            ref Result result
            )
        {
            result = FormatOps.PluginAbout(this, false);
            return ReturnCode.Ok;
        }

        ///////////////////////////////////////////////////////////////////////

        public override ReturnCode Options(
            Interpreter interpreter,
            ref Result result
            )
        {
            result = new StringList(DefineConstants.OptionList, false);
            return ReturnCode.Ok;
        }
        #endregion
    }
}
